@@ -1,7 +1,7 @@
 /*  bam_reheader.c -- reheader subcommand.
 
     Copyright (C) 2010 Broad Institute.
-    Copyright (C) 2012-2019 Genome Research Ltd.
+    Copyright (C) 2012-2019, 2025 Genome Research Ltd.
 
     Author: Heng Li <lh3@sanger.ac.uk>
 
@@ -30,6 +30,9 @@ DEALINGS IN THE SOFTWARE.  */
 #include <assert.h>
 #include <getopt.h>
 #include <unistd.h>
+#ifdef _WIN32
+#include <windows.h> // GetTempPath
+#endif
 
 #include "htslib/bgzf.h"
 #include "htslib/sam.h"
@@ -127,6 +130,11 @@ int cram_reheader(cram_fd *in, sam_hdr_t *h, const char *arg_list, int no_pg)
     if (!h)
         return ret;
 
+    // Match output version number with input file.
+    char vers[99];
+    sprintf(vers, "%d.%d", cram_major_vers(in), cram_minor_vers(in));
+    cram_set_option(out, CRAM_OPT_VERSION, vers);
+
     // Attempt to fill out a cram->refs[] array from @SQ headers
     sam_hdr_t *cram_h = sam_hdr_dup(h);
     if (!cram_h)
@@ -453,7 +461,21 @@ static sam_hdr_t* external_reheader(samFile* in, const char* external) {
         fprintf(stderr, "[%s] failed to read the header for '%s'.\n", __func__, in->fn);
         return NULL;
     }
-    char tmp_fn[] = "reheaderXXXXXX";
+    char tmp_fn[1024+15];
+    char *env = getenv("TMPDIR");
+#ifdef _WIN32
+    char tmp_path[MAX_PATH];
+    if (!env) {
+        int ret = GetTempPath(MAX_PATH, tmp_path);
+        if (!ret || ret > MAX_PATH)
+            strcpy(tmp_path, "./");
+        env = tmp_path;
+    }
+#else
+    if (!env)
+        env = "/tmp";
+#endif
+    snprintf(tmp_fn, sizeof(tmp_fn), "%s/reheaderXXXXXX", env);
     int tmp_fd = mkstemp(tmp_fn);
     if (tmp_fd < 0) {
         print_error_errno("reheader", "fail to open temp file '%s'", tmp_fn);
