@@ -1,6 +1,6 @@
 /*  consensus__pileup.h -- Pileup orientated data per consensus column
 
-    Copyright (C) 2013-2016, 2020-2021 Genome Research Ltd.
+    Copyright (C) 2013-2016, 2020-2022 Genome Research Ltd.
 
     Author: James Bonfied <jkb@sanger.ac.uk>
 
@@ -342,8 +342,9 @@ int pileup_loop(samFile *fp,
         hts_pos_t pos;
 
         r = seq_fetch(client_data, fp, h, &pnew->b);
+        //fprintf(stderr, "Fetch at %ld\n", pnew->b.core.pos);
         if (r < -1) {
-            fprintf(stderr, "bam_next_seq() failure.\n");
+            fprintf(stderr, "pileup_loop() seq_fetch failure.\n");
             goto error;
         }
 
@@ -548,8 +549,11 @@ int pileup_loop(samFile *fp,
             if (seq_init) {
                 int v;
                 v = seq_init(client_data, fp, h, p);
-                if (v == -1)
+                if (v == -1) {
+                    p->next = pfree;
+                    pfree = p;
                     goto error;
+                }
 
                 if (v == 1) {
                     /* Keep this seq */
@@ -602,3 +606,4 @@ int pileup_loop(samFile *fp,
 
     return ret;
 }
+
