@@ -1,7 +1,7 @@
 /*  phase.c -- phase subcommand.
 
     Copyright (C) 2011 Broad Institute.
-    Copyright (C) 2013-2016, 2019 Genome Research Ltd.
+    Copyright (C) 2013-2016, 2019, 2024 Genome Research Ltd.
 
     Author: Heng Li <lh3@sanger.ac.uk>
 
@@ -116,11 +116,21 @@ static void count1(int l, const uint8_t *seq, int *cnt)
 static int **count_all(int l, int vpos, nseq_t *hash)
 {
     khint_t k;
-    int i, j, **cnt;
-    uint8_t *seq;
+    int i, j, **cnt = NULL;
+    uint8_t *seq = NULL;
+    size_t cnt_sz = ((size_t)1) << l;
+    if (cnt_sz > SSIZE_MAX / sizeof(int) / vpos) {
+        errno = ENOMEM;
+        goto fail;
+    }
     seq = calloc(l, 1);
+    if (!seq) goto fail;
     cnt = calloc(vpos, sizeof(int*));
-    for (i = 0; i < vpos; ++i) cnt[i] = calloc(1<<l, sizeof(int));
+    if (!cnt) goto fail;
+    for (i = 0; i < vpos; ++i) {
+        cnt[i] = calloc(cnt_sz, sizeof(int));
+        if (!cnt[i]) goto fail;
+    }
     for (k = 0; k < kh_end(hash); ++k) {
         if (kh_exist(hash, k)) {
             frag_t *f = &kh_val(hash, k);
@@ -138,6 +148,15 @@ static int **count_all(int l, int vpos, nseq_t *hash)
     }
     free(seq);
     return cnt;
+ fail:
+    free(seq);
+    if (cnt) {
+        for (i = 0; i < vpos; i++)
+            free(cnt[i]);
+        free(cnt);
+    }
+    print_error_errno("phase", "Couldn't allocate memory for counts");
+    return NULL;
 }
 
 // phasing
@@ -413,6 +432,7 @@ static int phase(phaseg_t *g, const char *chr, int vpos, uint64_t *cns, nseq_t *
         printf("PS\t%s\t%d\t%d\n", chr, (int)(cns[0]>>32) + 1, (int)(cns[vpos-1]>>32) + 1);
         sitemask = calloc(vpos, 1);
         cnt = count_all(g->k, vpos, hash);
+        if (!cnt) return -1;
         path = dynaprog(g->k, vpos, cnt);
         for (i = 0; i < vpos; ++i) free(cnt[i]);
         free(cnt);
@@ -583,6 +603,7 @@ static int start_output(phaseg_t *g, int c, const char *middle, const htsFormat
 int main_phase(int argc, char *argv[])
 {
     int c, tid, pos, vpos = 0, n, lasttid = -1, max_vpos = 0, usage = 0;
+    int status = EXIT_SUCCESS;
     const bam_pileup1_t *plp;
     bam_plp_t iter;
     nseq_t *seqs;
@@ -596,6 +617,8 @@ int main_phase(int argc, char *argv[])
     sam_global_args ga = SAM_GLOBAL_ARGS_INIT;
     static const struct option lopts[] = {
         SAM_OPT_GLOBAL_OPTIONS('-', 0, 0, 0, 0, '-'),
+        {"min-BQ", required_argument, NULL, 'Q'},
+        {"min-bq", required_argument, NULL, 'Q'},
         {"no-PG", no_argument, NULL, 1},
         { NULL, 0, NULL, 0 }
     };
@@ -629,7 +652,8 @@ int main_phase(int argc, char *argv[])
         fprintf(stderr, "Options: -k INT    block length [%d]\n", g.k);
         fprintf(stderr, "         -b STR    prefix of BAMs to output [null]\n");
         fprintf(stderr, "         -q INT    min het phred-LOD [%d]\n", g.min_varLOD);
-        fprintf(stderr, "         -Q INT    min base quality in het calling [%d]\n", g.min_baseQ);
+        fprintf(stderr, "         -Q, --min-BQ INT\n"
+                        "                   min base quality in het calling [%d]\n", g.min_baseQ);
         fprintf(stderr, "         -D INT    max read depth [%d]\n", g.max_depth);
 //      fprintf(stderr, "         -l FILE   list of sites to phase [null]\n");
         fprintf(stderr, "         -F        do not attempt to fix chimeras\n");
@@ -785,6 +809,12 @@ int main_phase(int argc, char *argv[])
             return 1;
         }
     }
+
+    if (n < 0) {
+        print_error("phase", "error reading from '%s'", argv[optind]);
+        status = EXIT_FAILURE;
+    }
+
     sam_hdr_destroy(g.fp_hdr);
     bam_plp_destroy(iter);
     sam_close(g.fp);
@@ -809,5 +839,5 @@ int main_phase(int argc, char *argv[])
     }
     free(g.arg_list);
     sam_global_args_free(&ga);
-    return 0;
+    return status;
 }
