@@ -37,6 +37,7 @@ DEALINGS IN THE SOFTWARE.  */
 #include "htslib/faidx.h"
 #include "htslib/khash.h"
 #include "htslib/kstring.h"
+#include "htslib/hfile.h"
 #include "htslib/thread_pool.h"
 #include "htslib/hts_expr.h"
 #include "samtools.h"
@@ -85,8 +86,10 @@ typedef struct samview_settings {
     sam_hdr_t *header;
     samFile *in, *out, *un_out;
     int64_t count;
+    int64_t processed;
     int is_count;
     char *fn_in, *fn_idx_in, *fn_out, *fn_fai, *fn_un_out, *fn_out_idx, *fn_un_out_idx;
+    char *fn_counts;
     int fetch_pairs, nreglist;
     hts_reglist_t *reglist;
     int sanitize;
@@ -639,8 +642,9 @@ static int fetch_pairs_collect_mates(samview_settings_t *conf, hts_itr_multi_t *
             }
         }
 
-        if ( rec->core.mtid < 0 || (rec->core.flag & BAM_FMUNMAP) ) nunmap = 1;
-        if ( rec->core.mtid >= 0 ) {
+        if ( rec->core.mtid < 0 ) {
+            nunmap = 1;
+        } else {
             if (_reglist_push(&conf->reglist, &conf->nreglist, rec->core.mtid, rec->core.mpos,rec->core.mpos+1) != 0)
                 goto out;
         }
@@ -670,13 +674,17 @@ static int fetch_pairs_collect_mates(samview_settings_t *conf, hts_itr_multi_t *
              k = kh_get(names,mate_names,bam_get_qname(rec));
              if ( k != kh_end(mate_names) ) drop = 0;
         }
+        if ( drop )
+            continue;
         int p = 0;
-        if (!drop && (p=process_aln(conf->header, rec, conf))== 0) {
+        conf->processed++;
+        if ((p=process_aln(conf->header, rec, conf)) == 0) {
             if (adjust_tags(conf->header, rec, conf) != 0)
                 goto out;
             if (check_sam_write1(conf->out, conf->header, rec, conf->fn_out,
                                  &write_error) < 0)
                 goto out;
+            conf->count++;
         }
         if (p < 0)
             goto out;
@@ -706,6 +714,7 @@ static int fetch_pairs_collect_mates(samview_settings_t *conf, hts_itr_multi_t *
 // Common code for processing and writing a record
 static inline int process_one_record(samview_settings_t *conf, bam1_t *b,
                                      int *write_error) {
+    conf->processed++;
     if (conf->sanitize)
         if (bam_sanitize(conf->header, b, conf->sanitize) < 0)
             return -1;
@@ -814,6 +823,48 @@ static void aux_list_free(samview_settings_t *settings) {
         kh_destroy(aux_exists, settings->remove_tag);
 }
 
+static int write_counts_to_file(samview_settings_t *settings) {
+    kstring_t text = KS_INITIALIZE;
+    hFILE *outfile = NULL;
+    int ret = -1;
+    int r = ksprintf(&text,
+                     "{\n"
+                     "    \"records_processed\" : %"PRId64",\n"
+                     "    \"records_filter_accepted\" : %"PRId64",\n"
+                     "    \"records_filter_rejected\" : %"PRId64"\n"
+                     "}\n",
+                     settings->processed, settings->count,
+                     settings->processed - settings->count);
+    if (r < 0) {
+        print_error_errno("view", "failed to make read counts text");
+        goto out;
+    }
+    outfile = hopen(settings->fn_counts, "w");
+    if (!outfile) {
+        print_error_errno("view", "failed to open \"%s\"", settings->fn_counts);
+        goto out;
+    }
+    if (hwrite(outfile, ks_c_str(&text), ks_len(&text)) != ks_len(&text)) {
+        print_error_errno("view", "failed to write to \"%s\"",
+                          settings->fn_counts);
+        goto out;
+    }
+    r = hclose(outfile);
+    outfile = NULL;
+    if (r < 0) {
+        print_error_errno("view", "error on closing \"%s\"",
+                          settings->fn_counts);
+        goto out;
+    }
+    ret = 0;
+
+ out:
+    ks_free(&text);
+    if (outfile)
+        hclose_abruptly(outfile);
+    return ret;
+}
+
 int main_samview(int argc, char *argv[])
 {
     samview_settings_t settings;
@@ -871,6 +922,7 @@ int main_samview(int argc, char *argv[])
         {"remove-flags", required_argument, NULL, LONGOPT('r')},
         {"remove-tag", required_argument, NULL, 'x'},
         {"require-flags", required_argument, NULL, 'f'},
+        {"save-counts", required_argument, NULL, LONGOPT('c')},
         {"subsample", required_argument, NULL, LONGOPT('s')},
         {"subsample-seed", required_argument, NULL, LONGOPT('S')},
         {"tag", required_argument, NULL, 'd'},
@@ -883,6 +935,7 @@ int main_samview(int argc, char *argv[])
         {"use-index", no_argument, NULL, 'M'},
         {"with-header", no_argument, NULL, 'h'},
         {"sanitize", required_argument, NULL, 'z'},
+        {NULL, 0, NULL, 0}
     };
 
     /* parse command-line options */
@@ -898,6 +951,8 @@ int main_samview(int argc, char *argv[])
     opterr = 0;
 
     char *tmp;
+    int tmp_flag;
+
     while ((c = getopt_long(argc, argv,
                             "SbBcCt:h1Ho:O:q:f:F:G:ul:r:T:R:N:d:D:L:s:@:m:x:U:MXe:pPz:",
                             lopts, NULL)) >= 0) {
@@ -931,6 +986,7 @@ int main_samview(int argc, char *argv[])
             settings.count_rf |= SAM_SEQ;
             break;
         case 'c': settings.is_count = 1; break;
+        case LONGOPT('c'): settings.fn_counts = optarg; break;
         case 'S': break;
         case 'b': out_format = "b"; break;
         case 'C': out_format = "c"; break;
@@ -942,19 +998,47 @@ int main_samview(int argc, char *argv[])
         case 'U': settings.fn_un_out = strdup(optarg); break;
         case 'X': has_index_file = 1; break;
         case 'f':
-            settings.flag_on |= bam_str2flag(optarg);
+            tmp_flag = bam_str2flag(optarg);
+
+            if (tmp_flag < 0) {
+                print_error("view", "Unknown flag '%s'", optarg);
+                return 1;
+            }
+
+            settings.flag_on |= tmp_flag;
             settings.count_rf |= SAM_FLAG | SAM_RNEXT;
             break;
         case 'F':
-            settings.flag_off |= bam_str2flag(optarg);
+            tmp_flag = bam_str2flag(optarg);
+
+            if (tmp_flag < 0) {
+                print_error("view", "Unknown flag '%s'", optarg);
+                return 1;
+            }
+
+            settings.flag_off |= tmp_flag;
             settings.count_rf |= SAM_FLAG | SAM_RNEXT;
             break;
         case LONGOPT('g'):
-            settings.flag_anyon |= bam_str2flag(optarg);
+            tmp_flag = bam_str2flag(optarg);
+
+            if (tmp_flag < 0) {
+                print_error("view", "Unknown flag '%s'", optarg);
+                return 1;
+            }
+
+            settings.flag_anyon |= tmp_flag;
             settings.count_rf |= SAM_FLAG | SAM_RNEXT;
             break;
         case 'G':
-            settings.flag_alloff |= bam_str2flag(optarg);
+            tmp_flag = bam_str2flag(optarg);
+
+            if (tmp_flag < 0) {
+                print_error("view", "Unknown flag '%s'", optarg);
+                return 1;
+            }
+
+            settings.flag_alloff |= tmp_flag;
             settings.count_rf |= SAM_FLAG | SAM_RNEXT;
             break;
         case 'q':
@@ -1114,8 +1198,27 @@ int main_samview(int argc, char *argv[])
             }
             settings.count_rf = INT_MAX; // no way to know what we need
             break;
-        case LONGOPT('r'): settings.remove_flag |= bam_str2flag(optarg); break;
-        case LONGOPT('a'): settings.add_flag |= bam_str2flag(optarg); break;
+        case LONGOPT('r'):
+            tmp_flag = bam_str2flag(optarg);
+
+            if (tmp_flag < 0) {
+                print_error("view", "Unknown flag '%s'", optarg);
+                return 1;
+            }
+
+            settings.remove_flag |= tmp_flag;
+            break;
+
+        case LONGOPT('a'):
+            tmp_flag = bam_str2flag(optarg);
+
+            if (tmp_flag < 0) {
+                print_error("view", "Unknown flag '%s'", optarg);
+                return 1;
+            }
+
+            settings.add_flag |= tmp_flag;
+            break;
 
         case 'x':
             if (*optarg == '^') {
@@ -1406,6 +1509,11 @@ view_end:
         }
     }
 
+    if (settings.fn_counts && ret == 0) {
+        if (write_counts_to_file(&settings) < 0)
+            ret = EXIT_FAILURE;
+    }
+
     // close files, free and return
     if (settings.in) check_sam_close("view", settings.in, settings.fn_in, "standard input", &ret);
     if (settings.out) check_sam_close("view", settings.out, settings.fn_out, "standard output", &ret);
@@ -1474,12 +1582,14 @@ static int usage(FILE *fp, int exit_status, int is_long_help)
 "  -H, --header-only          Print SAM header only (no alignments)\n"
 "      --no-header            Print SAM alignment records only [default]\n"
 "  -c, --count                Print only the count of matching records\n"
+"      --save-counts FILE     Write counts of passed/failed records to FILE\n"
 "  -o, --output FILE          Write output to FILE [standard output]\n"
 "  -U, --unoutput FILE, --output-unselected FILE\n"
 "                             Output reads not selected by filters to FILE\n"
 "  -p, --unmap                Set flag to UNMAP on reads not selected\n"
 "                             then write to output file.\n"
 "  -P, --fetch-pairs          Retrieve complete pairs even when outside of region\n"
+"\n"
 "Input options:\n"
 "  -t, --fai-reference FILE   FILE listing reference names and lengths\n"
 "  -M, --use-index            Use index and multi-region iterator for regions\n"
@@ -1665,7 +1775,7 @@ int main_head(int argc, char *argv[])
     if (nrecords > 0) {
         b = bam_init1();
         uint64_t n;
-        int r;
+        int r = 0;
         for (n = 0; n < nrecords && (r = sam_read1(fp, hdr, b)) >= 0; n++) {
             if (sam_format1(hdr, b, &str) < 0) {
                 print_error_errno("head", "couldn't format record");
